from typing import Optional, List


def _custom_warning(message, category=UserWarning, filename='', lineno=-1,
                    *args):
    print(f"{filename}:{lineno}: {category.__name__}: {message}")


def delete_pystore(path: str, name: str,
                   libraries: Optional[List[str]] = None) -> None:
    """Delete libraries from pystore.

    Parameters
    ----------
    path : str
        path to pystore
    name : str
        name of the pystore
    libraries : Optional[List[str]], optional
        list of library names to delete, by default None which deletes
        all libraries
    """
    try:
        import pystore
    except ModuleNotFoundError as e:
        print("Please install `pystore`!")
        raise e
    print(f"Deleting pystore: '{name}' ...", end="")
    pystore.set_path(path)
    if libraries is None:
        pystore.delete_store(name)
        print(" Done!")
    else:
        store = pystore.store(name)
        for lib in libraries:
            print()
            store.delete_collection(lib)
            print(f" - deleted: {lib}")


def delete_arctic(connstr: str, name: str,
                  libraries: Optional[List[str]] = None) -> None:
    """Delete libraries from arctic database.

    Parameters
    ----------
    connstr : str
        connection string to the database
    name : str
        name of the database
    libraries : Optional[List[str]], optional
        list of library names to delete, by default None which deletes
        all libraries
    """
    try:
        import arctic
    except ModuleNotFoundError as e:
        print("Please install `arctic`!")
        raise e
    arc = arctic.Arctic(connstr)
    print(f"Deleting database: '{name}' ...")
    # get library names
    if libraries is None:
        libs = []
        for ilib in arc.list_libraries():
            if ilib.split(".")[0] == name:
                libs.append(ilib)
    else:
        libs = [name + "." + ilib for ilib in libraries]

    for l in libs:
        arc.delete_library(l)
        print(f" - deleted: {l}")
    print("... Done!")



import json
import os
import warnings
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd
from tqdm import tqdm

import pastas as ps
from pastas.io.pas import pastas_hook

from .util import _custom_warning

FrameorSeriesUnion = Union[pd.DataFrame, pd.Series]
warnings.showwarning = _custom_warning


class PastaStore:
    """Pastas project for managing pastas timeseries and models.

    Requires a Connector object to provide the interface to
    the database. Different Connectors are available, e.g.:

        - ArcticConnector for saving data to MongoDB using the Arctic module
        - PystoreConnector for saving data to disk using the Pystore module

    Parameters
    ----------
    name : str
        name of the project
    connector : Connector object
        object that provides the interface to the
        database, e.g. ArcticConnector (see pastastore.connectors)
    """

    def __init__(self, name: str, connector):
        """Initialize PastaStore for managing pastas timeseries and models.

        Parameters
        ----------
        name : str
            name of the project
        connector : Connector object
            object that provides the interface to the
            database
        """
        self.name = name
        self.conn = connector
        self._register_connector_methods()

    def _register_connector_methods(self):
        """Internal method for registering connector methods."""
        methods = [func for func in dir(self.conn) if
                   callable(getattr(self.conn, func)) and
                   not func.startswith("_")]
        for meth in methods:
            setattr(self, meth, getattr(self.conn, meth))

    def __repr__(self):
        """Representation string of the object."""
        return f"<PastaStore> {self.name}: \n - " + self.conn.__str__()

    def get_oseries_distances(self, names: Optional[Union[list, str]] = None) \
            -> FrameorSeriesUnion:
        """Method to obtain the distances in meters between the oseries.

        Parameters
        ----------
        names: str or list of str
            names of the oseries to calculate distances between

        Returns
        -------
        distances: pandas.DataFrame
            Pandas DataFrame with the distances between the oseries
        """
        oseries_df = self.conn.oseries
        other_df = self.conn.oseries

        names = self.conn._parse_names(names)

        xo = pd.to_numeric(oseries_df.loc[names, "x"])
        xt = pd.to_numeric(other_df.loc[:, "x"])
        yo = pd.to_numeric(oseries_df.loc[names, "y"])
        yt = pd.to_numeric(other_df.loc[:, "y"])

        xh, xi = np.meshgrid(xt, xo)
        yh, yi = np.meshgrid(yt, yo)

        distances = pd.DataFrame(np.sqrt((xh - xi) ** 2 + (yh - yi) ** 2),
                                 index=names, columns=other_df.index)

        return distances

    def get_nearest_oseries(self, names: Optional[Union[list, str]] = None,
                            n: int = 1) -> FrameorSeriesUnion:
        """Method to obtain the nearest (n) oseries.

        Parameters
        ----------
        names: str or list of str
            string or list of strings with the name(s) of the oseries
        n: int
            number of oseries to obtain

        Returns
        -------
        oseries:
            list with the names of the oseries.
        """

        distances = self.get_oseries_distances(names)

        data = pd.DataFrame(columns=np.arange(n))

        for series in distances.index:
            series = pd.Series(distances.loc[series].sort_values().index[:n],
                               name=series)
            data = data.append(series)
        return data

    def get_distances(self, oseries: Optional[Union[list, str]] = None,
                      stresses: Optional[Union[list, str]] = None,
                      kind: Optional[str] = None) -> FrameorSeriesUnion:
        """Method to obtain the distances in meters between the oseries and
        stresses.

        Parameters
        ----------
        oseries: str or list of str
            name(s) of the oseries
        stresses: str or list of str
            name(s) of the stresses
        kind: str
            string representing which kind of stresses to consider

        Returns
        -------
        distances: pandas.DataFrame
            Pandas DataFrame with the distances between the oseries (index)
            and the stresses (columns).
        """
        oseries_df = self.conn.oseries
        stresses_df = self.conn.stresses

        oseries = self.conn._parse_names(oseries)

        if stresses is None and kind is None:
            stresses = stresses_df.index
        elif stresses is not None and kind is not None:
            mask = stresses_df.kind == kind
            stresses = stresses_df.loc[stresses].loc[mask].index
        elif stresses is None:
            stresses = stresses_df.loc[stresses_df.kind == kind].index

        xo = pd.to_numeric(oseries_df.loc[oseries, "x"])
        xt = pd.to_numeric(stresses_df.loc[stresses, "x"])
        yo = pd.to_numeric(oseries_df.loc[oseries, "y"])
        yt = pd.to_numeric(stresses_df.loc[stresses, "y"])

        xh, xi = np.meshgrid(xt, xo)
        yh, yi = np.meshgrid(yt, yo)

        distances = pd.DataFrame(np.sqrt((xh - xi) ** 2 + (yh - yi) ** 2),
                                 index=oseries, columns=stresses)

        return distances

    def get_nearest_stresses(self, oseries: Optional[Union[list, str]] = None,
                             stresses: Optional[Union[list, str]] = None,
                             kind: Optional[str] = None, n: int = 1) -> \
            FrameorSeriesUnion:
        """Method to obtain the nearest (n) stresses of a specific kind.

        Parameters
        ----------
        oseries: str
            string with the name of the oseries
        stresses: str or list of str
            string with the name of the stresses
        kind:
            string with the name of the stresses
        n: int
            number of stresses to obtain

        Returns
        -------
        stresses:
            list with the names of the stresses.
        """

        distances = self.get_distances(oseries, stresses, kind)

        data = pd.DataFrame(columns=np.arange(n))

        for series in distances.index:
            series = pd.Series(distances.loc[series].sort_values().index[:n],
                               name=series)
            data = data.append(series)
        return data

    def get_tmin_tmax(self, libname, names=None, progressbar=False):
        """Get tmin and tmax for timeseries.

        Parameters
        ----------
        libname : str
            name of the library containing the timeseries
            ('oseries' or 'stresses')
        names : str, list of str, or None, optional
            names of the timeseries, by default None which
            uses all the timeseries in the library
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        tmintmax : pd.dataframe
            Dataframe containing tmin and tmax per timeseries
        """

        names = self.conn._parse_names(names, libname=libname)
        tmintmax = pd.DataFrame(index=names, columns=["tmin", "tmax"],
                                dtype='datetime64[ns]')
        desc = f"Get tmin/tmax {libname}"
        for n in (tqdm(names, desc=desc) if progressbar else names):
            if libname == "oseries":
                s = self.conn.get_oseries(n)
            else:
                s = self.conn.get_stresses(n)
            tmintmax.loc[n, "tmin"] = s.first_valid_index()
            tmintmax.loc[n, "tmax"] = s.last_valid_index()
        return tmintmax

    def get_parameters(self, parameters=None, modelnames=None,
                       param_value="optimal", progressbar=False):
        """Get model parameters. NaN-values are returned when the parameters
        are not present in the model or the model is not optimized.

        Parameters
        ----------
        parameters : list of str, optional
            names of the parameters, by default None which uses all
            parameters from each model
        modelnames : str or list of str, optional
            name(s) of model(s), by default None in which case all models
            are used
        param_value : str, optional
            which column to use from the model parameters dataframe, by
            default "optimal" which retrieves the optimized parameters.
        progressbar : bool, optional
            show progressbar, default is False

        Returns
        -------
        p : pandas.DataFrame
            DataFrame containing the parameters (columns) per model (rows)
        """
        modelnames = self.conn._parse_names(modelnames, libname="models")

        # create dataframe for results
        p = pd.DataFrame(index=modelnames, columns=parameters)

        # loop through model names and store results
        desc = "Get model parameters"
        for mlname in (tqdm(modelnames, desc=desc) if progressbar else modelnames):
            mldict = self.get_models(mlname, return_dict=True,
                                     progressbar=False)
            if parameters is None:
                pindex = mldict["parameters"].index
            else:
                pindex = parameters

            for c in pindex:
                p.loc[mlname, c] = \
                    mldict["parameters"].loc[c, param_value]

        p = p.squeeze()
        return p.astype(float)

    def get_statistics(self, statistics, modelnames=None, progressbar=False,
                       **kwargs):
        """Get model statistics.

        Parameters
        ----------
        statistics : list of str
            list of statistics to calculate, e.g. ["evp", "rsq", "rmse"], for
            a full list see `pastas.modelstats.Statistics.ops`.
        modelnames : list of str, optional
            modelnames to calculates statistics for, by default None, which
            uses all models in the store
        progressbar : bool, optional
            show progressbar, by default False
        **kwargs
            any arguments that can be passed to the methods for calculating
            statistics

        Returns
        -------
        s : pandas.DataFrame
        """

        modelnames = self.conn._parse_names(modelnames, libname="models")

        # create dataframe for results
        s = pd.DataFrame(index=modelnames, columns=statistics)

        # loop through model names
        desc = "Get model statistics"
        for mlname in (tqdm(modelnames, desc=desc) if progressbar else modelnames):
            ml = self.get_models(mlname, progressbar=False)
            for stat in statistics:
                value = ml.stats.__getattribute__(stat)(**kwargs)
                s.loc[mlname, stat] = value

        s = s.squeeze()
        return s.astype(float)

    def create_model(self, name: str, add_recharge: bool = True) -> ps.Model:
        """Create a pastas Model.

        Parameters
        ----------
        name : str
            name of the oseries to create a model for
        add_recharge : bool, optional
            add recharge to the model by looking for the closest
            precipitation and evaporation timeseries in the stresses
            library, by default True

        Returns
        -------
        pastas.Model
            model for the oseries

        Raises
        ------
        KeyError
            if data is stored as dataframe and no column is provided
        ValueError
            if timeseries is empty
        """
        # get oseries metadata
        meta = self.conn.get_metadata("oseries", name, as_frame=False)
        ts = self.conn.get_oseries(name)

        # convert to Timeseries and create model
        if not ts.dropna().empty:
            ts = ps.TimeSeries(ts, name=name, settings="oseries",
                               metadata=meta)
            ml = ps.Model(ts, name=name, metadata=meta)

            if add_recharge:
                self.add_recharge(ml)
            return ml
        else:
            raise ValueError("Empty timeseries!")

    def create_models_bulk(self, oseries: Optional[Union[list, str]] = None,
                           add_recharge: bool = True, store: bool = True,
                           solve: bool = False, progressbar: bool = True,
                           return_models: bool = False,
                           ignore_errors: bool = False,
                           **kwargs) -> Union[Tuple[dict, dict], dict]:
        """Bulk creation of pastas models.

        Parameters
        ----------
        oseries : list of str, optional
            names of oseries to create models for, by default None,
            which creates models for all oseries
        add_recharge : bool, optional
            add recharge to the models based on closest
            precipitation and evaporation timeseries, by default True
        store : bool, optional
            store the model, by default True
        solve : bool, optional
            solve the model, by default False
        progressbar : bool, optional
            show progressbar, by default True
        return_models : bool, optional
            if True, return a list of models, by default False
        ignore_errors : bool, optional
            ignore errors while creating models, by default False

        Returns
        -------
        models : dict, if return_models is True
            dictionary of models
        errors : list, always returned
            list of model names that could not be created
        """
        if oseries is None:
            oseries = self.conn.oseries.index
        elif isinstance(oseries, str):
            oseries = [oseries]

        models = {}
        errors = {}
        desc = "Bulk creation models"
        for o in (tqdm(oseries, desc=desc) if progressbar else oseries):
            try:
                iml = self.create_model(o, add_recharge=add_recharge)
            except Exception as e:
                if ignore_errors:
                    errors[o] = e
                    continue
                else:
                    raise e
            if solve:
                iml.solve(**kwargs)
            if store:
                self.conn.add_model(iml, overwrite=True)
            if return_models:
                models[o] = iml
        if len(errors) > 0:
            print("Warning! Errors occurred while creating models!")
        if return_models:
            return models, errors
        else:
            return errors

    def add_recharge(self, ml: ps.Model, rfunc=ps.Gamma) -> None:
        """Add recharge to a pastas model.

        Uses closest precipitation and evaporation timeseries in database.
        These are assumed to be labeled with kind = 'prec' or 'evap'.

        Parameters
        ----------
        ml : pastas.Model
            pastas.Model object
        rfunc : pastas.rfunc, optional
            response function to use for recharge in model,
            by default ps.Gamma (for different response functions, see
            pastas documentation)
        """
        # get nearest prec and evap stns
        names = []
        for var in ("prec", "evap"):
            try:
                name = self.get_nearest_stresses(
                    ml.oseries.name, kind=var).iloc[0, 0]
            except AttributeError:
                msg = "No precipitation or evaporation timeseries found!"
                raise Exception(msg)
            names.append(name)
        if len(names) == 0:
            msg = "No precipitation or evaporation timeseries found!"
            raise Exception(msg)

        # get data
        tsdict = self.conn.get_stresses(names)
        stresses = []
        for k, s in tsdict.items():
            metadata = self.conn.get_metadata("stresses", k, as_frame=False)
            stresses.append(ps.TimeSeries(s, name=k, metadata=metadata))

        # add recharge to model
        rch = ps.StressModel2(stresses, rfunc, name="recharge",
                              metadata=[i.metadata for i in stresses],
                              settings=("prec", "evap"))
        ml.add_stressmodel(rch)

    def solve_models(self, mls: Optional[Union[ps.Model, list, str]] = None,
                     report: bool = False, ignore_solve_errors: bool = False,
                     store_result: bool = True, progressbar: bool = True,
                     **kwargs) -> None:
        """Solves the models in the store.

        Parameters
        ----------
        mls : list of str, optional
            list of model names, if None all models in the project
            are solved.
        report : boolean, optional
            determines if a report is printed when the model is solved,
            default is False
        ignore_solve_errors : boolean, optional
            if True, errors emerging from the solve method are ignored,
            default is False which will raise an exception when a model
            cannot be optimized
        store_result : bool, optional
            if True save optimized models, default is True
        progressbar : bool, optional
            show progressbar, default is True
        **kwargs :
            arguments are passed to the solve method.
        """
        if mls is None:
            mls = self.conn.models
        elif isinstance(mls, ps.Model):
            mls = [mls.name]

        desc = "Solving models"
        for ml_name in (tqdm(mls, desc=desc) if progressbar else mls):
            ml = self.conn.get_models(ml_name)

            m_kwargs = {}
            for key, value in kwargs.items():
                if isinstance(value, pd.Series):
                    m_kwargs[key] = value.loc[ml_name]
                else:
                    m_kwargs[key] = value
            # Convert timestamps
            for tstamp in ["tmin", "tmax"]:
                if tstamp in m_kwargs:
                    m_kwargs[tstamp] = pd.Timestamp(m_kwargs[tstamp])

            try:
                ml.solve(report=report, **m_kwargs)
                if store_result:
                    self.conn.add_model(ml, overwrite=True)
            except Exception as e:
                if ignore_solve_errors:
                    warning = "solve error ignored for -> {}".format(ml.name)
                    ps.logger.warning(warning)
                else:
                    raise e

    def model_results(self, mls: Optional[Union[ps.Model, list, str]] = None,
                      progressbar: bool = True):  # pragma: no cover
        """Get pastas model results.

        Parameters
        ----------
        mls : list of str, optional
            list of model names, by default None which means results for
            all models will be calculated
        progressbar : bool, optional
            show progressbar, by default True

        Returns
        -------
        results : pd.DataFrame
            dataframe containing parameters and other statistics
            for each model

        Raises
        ------
        ModuleNotFoundError
            if the art_tools module is not available
        """
        try:
            from art_tools import (
                pastas_get_model_results)
        except:
            raise ModuleNotFoundError(
                "You need 'art_tools' to use this method!")

        if mls is None:
            mls = self.conn.models
        elif isinstance(mls, ps.Model):
            mls = [mls.name]

        results_list = []
        desc = "Get model results"
        for mlname in (tqdm(mls, desc=desc) if progressbar else mls):
            try:
                iml = self.conn.get_models(mlname)
            except Exception as e:
                print("{1}: '{0}' could not be parsed!".format(mlname, e))
                continue
            iresults = pastas_get_model_results(
                iml, par_selection='all', stats=('evp',), stderrors=True)
            results_list.append(iresults)

        return pd.concat(results_list, axis=1).transpose()

    def to_zip(self, fname: str, overwrite=False, progressbar: bool = True):
        """Write data to zipfile.

        Parameters
        ----------
        fname : str
            name of zipfile
        overwrite : bool, optional
            if True, overwrite existing file
        progressbar : bool, optional
            show progressbar, by default True
        """
        from zipfile import ZipFile, ZIP_DEFLATED
        if os.path.exists(fname) and not overwrite:
            raise FileExistsError("File already exists! "
                                  "Use 'overwrite=True' to "
                                  "force writing file.")
        elif os.path.exists(fname):
            warnings.warn(f"Overwriting file '{os.path.basename(fname)}'")

        with ZipFile(fname, "w", compression=ZIP_DEFLATED) as archive:
            # oseries
            self.conn._series_to_archive(archive, "oseries",
                                         progressbar=progressbar)
            # stresses
            self.conn._series_to_archive(archive, "stresses",
                                         progressbar=progressbar)
            # models
            self.conn._models_to_archive(archive, progressbar=progressbar)

    @classmethod
    def from_zip(cls, fname: str, conn, storename: Optional[str] = None):
        """Load PastaStore from zipfile.

        Parameters
        ----------
        fname : str
            pathname of zipfile
        conn : Connector object
            connector for storing loaded data
        storename : str, optional
            name of the PastaStore, by default None, which
            defaults to the name of the Connector.

        Returns
        -------
        pastastore.PastaStore
            return PastaStore containing data from zipfile
        """
        from zipfile import ZipFile
        with ZipFile(fname, "r") as archive:
            namelist = [fi for fi in archive.namelist()
                        if not fi.endswith("_meta.json")]
            for f in namelist:
                libname, fjson = os.path.split(f)
                if libname in ["stresses", "oseries"]:
                    s = pd.read_json(archive.open(f),
                                     orient="columns")
                    meta = json.load(archive.open(
                        f.replace(".json", "_meta.json")))
                    conn._add_series(libname, s, fjson.split(".")[0],
                                     metadata=meta)
                elif libname in ["models"]:
                    ml = json.load(archive.open(f), object_hook=pastas_hook)
                    conn.add_model(ml)
        if storename is None:
            storename = conn.name
        return cls(storename, conn)

    @ property
    def oseries(self):
        return self.conn.oseries

    @ property
    def stresses(self):
        return self.conn.stresses

    @ property
    def models(self):
        return self.conn.models



import json
from abc import ABC, abstractmethod, abstractproperty
from collections.abc import Iterable
from typing import Optional, Union
import warnings

import pandas as pd
from tqdm import tqdm

import pastas as ps
from pastas import Model
from pastas.io.pas import PastasEncoder

from .util import _custom_warning

FrameorSeriesUnion = Union[pd.DataFrame, pd.Series]
warnings.showwarning = _custom_warning


class BaseConnector(ABC):  # pragma: no cover
    """Metaclass for connecting to data management sources.

    For example, MongoDB through Arctic, Pystore, or other databases.
    Create your own connection to a data source by writing a a class
    that inherits from this BaseConnector. Your class has to override
    each method and property.
    """
    _default_library_names = ["oseries", "stresses", "models"]

    @abstractmethod
    def get_library(self, libname: str):
        """Get library handle.

        Parameters
        ----------
        libname : str,
            name of the library
        """
        pass

    @abstractmethod
    def add_oseries(self, series: FrameorSeriesUnion, name: str,
                    metadata: Union[dict, None] = None,
                    overwrite: bool = False, **kwargs) -> None:
        """Add oseries.

        Parameters
        ----------
        series : FrameorSeriesUnion
            pandas.Series or pandas.DataFrame to add
        name : str
            name of the series
        metadata : Optional[dict], optional
            dictionary containing metadata, by default None
        overwrite: bool, optional
            overwrite existing dataset with the same name,
            by default False
        """
        pass

    @abstractmethod
    def add_stress(self, series: FrameorSeriesUnion, name: str, kind: str,
                   metadata: Optional[dict] = None,
                   overwrite: bool = False, **kwargs) -> None:
        """Add stress.

        Parameters
        ----------
        series : FrameorSeriesUnion
            pandas.Series or pandas.DataFrame to add
        name : str
            name of the series
        kind : str
            label specifying type of stress (i.e. 'prec' or 'evap')
        metadata : Optional[dict], optional
            dictionary containing metadata, by default None
        overwrite: bool, optional
            overwrite existing dataset with the same name,
            by default False
        """
        pass

    @abstractmethod
    def add_model(self, ml: Model, overwrite: bool = False, **kwargs) -> None:
        """Add model.

        Parameters
        ----------
        ml : Model
            pastas.Model
        """
        pass

    @abstractmethod
    def del_models(self, names: Union[list, str]) -> None:
        """Delete model.

        Parameters
        ----------
        names : Union[list, str]
            str or list of str of model names to delete
        """
        pass

    @abstractmethod
    def del_oseries(self, names: Union[list, str]) -> None:
        """Delete oseries.

        Parameters
        ----------
        names : Union[list, str]
            str or list of str of oseries names to delete
        """
        pass

    @abstractmethod
    def del_stress(self, names: Union[list, str]) -> None:
        """Delete stresses.

        Parameters
        ----------
        names : Union[list, str]
            str or list of str of stresses to delete
        """
        pass

    @abstractmethod
    def get_metadata(self, libname: str, names: Union[list, str],
                     progressbar: bool = False, as_frame: bool = True) -> Union[pd.DataFrame, dict]:
        """Get metadata for oseries or stress.

        Parameters
        ----------
        libname : str
            name of library
        names : Union[list, str]
            str or list of str of series to get metadata for
        progressbar : bool, optional
            show progressbar, by default False
        as_frame : bool, optional
            return as dataframe, by default True

        Returns
        -------
        Union[pd.DataFrame, dict]
            dictionary or pandas.DataFrame depending on value of `as_frame`.
        """
        pass

    @abstractmethod
    def get_oseries(self, names: Union[list, str],
                    progressbar: bool = False) -> FrameorSeriesUnion:
        """Get oseries.

        Parameters
        ----------
        names : Union[list, str]
            str or list of str of names of oseries to retrieve
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        dict, pandas.DataFrame
            return dictionary containing data if multiple names are passed,
            else return pandas.DataFrame or pandas.Series
        """
        pass

    @abstractmethod
    def get_stresses(self, names: Union[list, str],
                     progressbar: bool = False) -> FrameorSeriesUnion:
        """Get stresses.

        Parameters
        ----------
        names : Union[list, str]
            str or list of str of names of stresses to retrieve
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        dict, pandas.DataFrame
            return dictionary containing data if multiple names are passed,
            else return pandas.DataFrame or pandas.Series
        """
        pass

    @abstractmethod
    def get_models(self, names: Union[list, str], progressbar: bool = False,
                   **kwargs) -> Union[Model, dict]:
        """Get models.

        Parameters
        ----------
        names : Union[list, str]
            str or list of str of models to retrieve
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        Union[Model, dict]
            return pastas.Model if only one name is passed, else return dict
        """
        pass

    @abstractproperty
    def oseries(self):
        """Dataframe containing oseries overview."""
        pass

    @abstractproperty
    def stresses(self):
        """Dataframe containing stresses overview."""
        pass

    @abstractproperty
    def models(self):
        """List of model names."""
        pass


class ConnectorUtil:
    """Mix-in class for general Connector helper functions."""

    def _parse_names(self, names: Optional[Union[list, str]] = None,
                     libname: Optional[str] = "oseries") -> list:
        """Internal method to parse names kwarg, returns iterable with name(s).

        Parameters
        ----------
        names : Union[list, str], optional
            str or list of str or None (retrieves all names)
        libname : str, optional
            name of library, default is 'oseries'

        Returns
        -------
        list
            list of names
        """
        if names is None:
            if libname == "oseries":
                return getattr(self, "oseries").index.to_list()
            elif libname == "stresses":
                return getattr(self, "stresses").index.to_list()
            elif libname == "models":
                return getattr(self, "models")
            else:
                raise ValueError(f"No library '{libname}'!")
        elif isinstance(names, str):
            return [names]
        elif isinstance(names, Iterable):
            return names
        else:
            raise NotImplementedError(f"Cannot parse 'names': {names}")

    @staticmethod
    def _meta_list_to_frame(metalist: list, names: list):
        """Convert list of metadata dictionaries to DataFrame.

        Parameters
        ----------
        metalist : list
            list of metadata dictionaries
        names : list
            list of names corresponding to data in metalist

        Returns
        -------
        pandas.DataFrame
            DataFrame containing overview of metadata
        """
        # convert to dataframe
        if len(metalist) > 1:
            meta = pd.DataFrame(metalist)
            if len({"x", "y"}.difference(meta.columns)) == 0:
                meta["x"] = meta["x"].astype(float)
                meta["y"] = meta["y"].astype(float)
                # meta = gpd.GeoDataFrame(meta, geometry=[Point(
                #     (s["x"], s["y"])) for i, s in meta.iterrows()])
        elif len(metalist) == 1:
            meta = pd.DataFrame(metalist)
        elif len(metalist) == 0:
            meta = pd.DataFrame()
        if "name" in meta.columns:
            meta.set_index("name", inplace=True)
        else:
            meta.index = names
        return meta

    def _parse_model_dict(self, mdict: dict):
        """Internal method to parse dictionary describing pastas models.

        Parameters
        ----------
        mdict : dict
            dictionary describing pastas.Model

        Returns
        -------
        ml : pastas.Model
            timeseries analysis model
        """
        if 'series' not in mdict['oseries']:
            name = mdict["oseries"]['name']
            if name not in self.oseries.index:
                msg = 'oseries {} not present in project'.format(name)
                raise LookupError(msg)
            mdict['oseries']['series'] = self.get_oseries(name)
        for ts in mdict["stressmodels"].values():
            if "stress" in ts.keys():
                for stress in ts["stress"]:
                    if 'series' not in stress:
                        name = stress['name']
                        if name in self.stresses.index:
                            stress['series'] = self.get_stresses(name)
                        else:
                            msg = "stress '{}' not present in project".format(
                                name)
                            raise KeyError(msg)
        try:
            # pastas>=0.15.0
            ml = ps.io.base._load_model(mdict)
        except AttributeError:
            # pastas<0.15.0
            ml = ps.io.base.load_model(mdict)
        return ml

    @staticmethod
    def _validate_input_series(series):
        """check if series is pandas.DataFrame or pandas.Series.

        Parameters
        ----------
        series : object
            object to validate

        Raises
        ------
        TypeError
            if object is not of type pandas.DataFrame or pandas.Series
        """
        if not (isinstance(series, pd.DataFrame) or
                isinstance(series, pd.Series)):
            raise TypeError("Please provide pandas.DataFrame"
                            " or pandas.Series!")
        if isinstance(series, pd.DataFrame):
            if series.columns.size > 1:
                raise ValueError("Only DataFrames with one "
                                 "column are supported!")

    @staticmethod
    def _set_series_name(series, name):
        """Set series name to match user defined name in store.

        Parameters
        ----------
        series : pandas.Series or pandas.DataFrame
            set name for this timeseries
        name : str
            name of the timeseries (used in the pastastore)
        """
        if isinstance(series, pd.Series):
            series.name = name
        if isinstance(series, pd.DataFrame):
            series.columns = [name]
        return series

    @staticmethod
    def _check_model_series_for_store(ml):
        if isinstance(ml, ps.Model):
            series_names = [istress.series.name for sm in
                            ml.stressmodels.values() for istress in sm.stress]
        elif isinstance(ml, dict):
            series_names = [istress["name"] for sm in
                            ml["stressmodels"].values()
                            for istress in sm["stress"]]
        else:
            raise TypeError("Expected pastas.Model or dict!")
        if len(series_names) - len(set(series_names)) > 0:
            msg = ("There are multiple stresses series with the same name! "
                   "Each series name must be unique for the PastaStore!")
            raise ValueError(msg)

    def _check_oseries_in_store(self, ml: Union[ps.Model, dict]):
        """Internal method, check if Model oseries are contained in PastaStore.

        Parameters
        ----------
        ml : Union[ps.Model, dict]
            pastas Model
        """
        if isinstance(ml, ps.Model):
            name = ml.oseries.name
        elif isinstance(ml, dict):
            name = ml["oseries"]["name"]
        else:
            raise TypeError("Expected pastas.Model or dict!")
        if name not in self.oseries.index:
            msg = (f"Cannot add model because oseries '{name}' "
                   "is not contained in store.")
            raise LookupError(msg)
            # if s is not None:
            #     warnings.warn(f"Adding '{name}' to 'oseries' store!")
            #     s.index.name = None
            #     self.add_oseries(s, name, metadata=meta)

    def _check_stresses_in_store(self, ml: Union[ps.Model, dict]):
        """Internal method, check if stresses timeseries are contained in
        PastaStore.

        Parameters
        ----------
        ml : Union[ps.Model, dict]
            pastas Model
        """
        if isinstance(ml, ps.Model):
            for sm in ml.stressmodels.values():
                for s in sm.stress:
                    if s.name not in self.stresses.index:
                        msg = (f"Cannot add model because stress '{s.name}' "
                               "is not contained in store.")
                        raise LookupError(msg)
                        # warnings.warn(f"Timeseries '{s.name}' "
                        #               "is not contained in store.")
                        # warnings.warn(
                        #     f"Adding '{s.name}' to 'stresses' store!")
                        # ss = s.series
                        # ss.index.name = None
                        # if "kind" not in s.metadata:
                        #     msg = ("Stress metadata dictionary must define"
                        #            " 'kind'!")
                        #     raise AttributeError(msg)
                        # self.add_stress(ss, s.name,
                        #                 metadata=s.metadata,
                        #                 kind=s.metadata["kind"])
        elif isinstance(ml, dict):
            for sm in ml["stressmodels"].values():
                for s in sm["stress"]:
                    if s["name"] not in self.stresses.index:
                        msg = (f"Cannot add model because stress '{s.name}' "
                               "is not contained in store.")
                        raise LookupError(msg)
                        # if "series" in s:
                        #     warnings.warn(
                        #         f"Adding '{s['name']}' to 'stresses' store!")
                        #     ss = s["series"]
                        #     ss.index.name = None
                        #     if "kind" not in s["metadata"]:
                        #         msg = ("Stress metadata dictionary must define"
                        #                " 'kind'!")
                        #         raise AttributeError(msg)
                        #     self.add_stress(ss, s["name"],
                        #                     metadata=s["metadata"],
                        #                     kind=s["metadata"]["kind"])
        else:
            raise TypeError("Expected pastas.Model or dict!")

    def _series_to_archive(self, archive, libname: str,
                           names: Optional[Union[list, str]] = None,
                           progressbar: bool = True):
        """Internal method for writing DataFrame or Series to zipfile.

        Parameters
        ----------
        archive : zipfile.ZipFile
            reference to an archive to write data to
        libname : str
            name of the library to write to zipfile
        names : str or list of str, optional
            names of the timeseries to write to archive, by default None,
            which writes all timeseries to archive
        progressbar : bool, optional
            show progressbar, by default True
        """
        names = self._parse_names(names, libname=libname)
        for n in (tqdm(names, desc=libname) if progressbar else names):
            s = self._get_series(libname, n, progressbar=False)
            if isinstance(s, pd.Series):
                s = s.to_frame()
            sjson = s.to_json(orient="columns")
            archive.writestr(f"{libname}/{n}.json", sjson)

            meta = self.get_metadata(libname, n, as_frame=False)
            meta_json = json.dumps(meta, cls=PastasEncoder, indent=4)
            archive.writestr(f"{libname}/{n}_meta.json", meta_json)

    def _models_to_archive(self, archive, names=None, progressbar=True):
        """Internal method for writing pastas.Model to zipfile.

        Parameters
        ----------
        archive : zipfile.ZipFile
            reference to an archive to write data to
        names : str or list of str, optional
            names of the models to write to archive, by default None,
            which writes all models to archive
        progressbar : bool, optional
            show progressbar, by default True
        """
        names = self._parse_names(names, libname="models")
        for n in (tqdm(names, desc="models") if progressbar else names):
            m = self.get_models(n, return_dict=True)
            jsondict = json.dumps(m, cls=PastasEncoder, indent=4)
            archive.writestr(f"models/{n}.pas", jsondict)



import functools
import json
import warnings
from copy import deepcopy
from importlib import import_module
from typing import Optional, Union

import pandas as pd
from tqdm import tqdm

import pastas as ps
from pastas import Model
from pastas.io.pas import PastasEncoder

from .base import BaseConnector, ConnectorUtil
from .util import _custom_warning

FrameorSeriesUnion = Union[pd.DataFrame, pd.Series]
warnings.showwarning = _custom_warning


class ArcticConnector(BaseConnector, ConnectorUtil):
    """Object to serve as the interface between MongoDB and Python using the
    Arctic module. Provides all the methods to read, write, or delete data from
    the database.

    Create an ArcticConnector object that connects to a
    running MongoDB database via Arctic.

    Parameters
    ----------
    connstr : str
        connection string
    projectname : str
        name of the project
    library_map: dict, optional
        dictionary containing the default library names as
        keys ('oseries', 'stresses', 'models') and the user
        specified library names as corresponding values.
        Allows user defined library names.
    """

    conn_type = "arctic"

    def __init__(self, name: str, connstr: str,
                 library_map: Optional[dict] = None):
        """Create an ArcticConnector object that connects to a running MongoDB
        database via Arctic.

        Parameters
        ----------
        connstr : str
            connection string
        name : str
            name of the project
        library_map: dict, optional
            dictionary containing the default library names as
            keys ('oseries', 'stresses', 'models') and the user
            specified library names as corresponding values.
            Allows user defined library names.
        """
        try:
            import arctic
        except ModuleNotFoundError as e:
            print("Please install arctic (also requires "
                  "a MongoDB instance running somewhere, e.g. "
                  "MongoDB Community: \n"
                  "https://docs.mongodb.com/manual/administration"
                  "/install-community/)!")
            raise e
        self.connstr = connstr
        self.name = name

        self.libs: dict = {}
        self.arc = arctic.Arctic(connstr)
        self._initialize(library_map)

    def __repr__(self):
        """Representation string of the object."""
        noseries = len(self.get_library("oseries").list_symbols())
        nstresses = len(self.get_library("stresses").list_symbols())
        nmodels = len(self.get_library("models").list_symbols())
        return ("<ArcticConnector object> '{0}': {1} oseries, "
                "{2} stresses, {3} models".format(
                    self.name, noseries, nstresses, nmodels))

    def _initialize(self, library_map: Optional[dict]) -> None:
        """Internal method to initalize the libraries."""
        if library_map is None:
            libmap = {i: i for i in self._default_library_names}
        else:
            libmap = library_map

        self.library_map = libmap

        for libname in libmap.values():
            if self._library_name(libname) not in self.arc.list_libraries():
                self.arc.initialize_library(self._library_name(libname))
            self.libs[libname] = self.get_library(libname)

    def _library_name(self, libname: str) -> str:
        """Internal method to get full library name according to Arctic."""
        return ".".join([self.name, libname])

    def get_library(self, libname: str):
        """Get Arctic library handle.

        Parameters
        ----------
        libname : str
            name of the library

        Returns
        -------
        Arctic.library handle
            handle to the library
        """
        # get custom library name if necessary
        real_libname = self.library_map[libname]

        # get library handle
        lib = self.arc.get_library(self._library_name(real_libname))
        return lib

    def _add_series(self, libname: str, series: FrameorSeriesUnion, name: str,
                    metadata: Optional[dict] = None,
                    overwrite: bool = False) -> None:
        """Internal method to add series to database.

        Parameters
        ----------
        libname : str
            name of the library to add the series to
        series : pandas.Series or pandas.DataFrame
            data to add
        name : str
            name of the timeseries
        metadata : dict, optional
            dictionary containing metadata, by default None
        overwrite : bool, optional
            overwrite existing dataset with the same name,
            by default False

        Raises
        ------
        Exception
            if overwrite is False and name is already in the database
            raises an Exception.
        """
        self._validate_input_series(series)
        series = self._set_series_name(series, name)
        lib = self.get_library(libname)
        if name not in lib.list_symbols() or overwrite:
            lib.write(name, series, metadata=metadata)
            self._clear_cache(libname)
        else:
            raise Exception("Item with name '{0}' already"
                            " in '{1}' library!".format(name, libname))

    def add_oseries(self, series: FrameorSeriesUnion, name: str,
                    metadata: Optional[dict] = None,
                    overwrite: bool = False) -> None:
        """Add oseries to the database.

        Parameters
        ----------
        series : pandas.Series or pandas.DataFrame
            data to add
        name : str
            name of the timeseries
        metadata : dict, optional
            dictionary containing metadata, by default None
        overwrite : bool, optional
            overwrite existing dataset with the same name,
            by default False
        """
        if isinstance(series, pd.DataFrame) and len(series.columns) > 1:
            if metadata is None:
                print("Data contains multiple columns, "
                      "assuming values in column 0!")
                metadata = {"value_col": 0}
            elif not "value_col" in metadata.keys():
                print("Data contains multiple columns, "
                      "assuming values in column 0!")
                metadata["value_col"] = 0

        self._add_series("oseries", series, name=name,
                         metadata=metadata, overwrite=overwrite)

    def add_stress(self, series: FrameorSeriesUnion, name: str, kind: str,
                   metadata: Optional[dict] = None,
                   overwrite: bool = False) -> None:
        """Add stress to the database.

        Parameters
        ----------
        series : pandas.Series or pandas.DataFrame
            data to add
        name : str
            name of the timeseries
        kind : str
            category to identify type of stress, this label is added to the
            metadata dictionary.
        metadata : dict, optional
            dictionary containing metadata, by default None. Also used to
            point to column containing timeseries if DataFrame is passed
            using the "value_col" key.
        overwrite : bool, optional
            overwrite existing dataset with the same name,
            by default False
        """
        if metadata is None:
            metadata = {}

        if isinstance(series, pd.DataFrame) and len(series.columns) > 1:
            print("Data contains multiple columns, "
                  "assuming values in column 0!")
            metadata["value_col"] = 0

        metadata["kind"] = kind
        self._add_series("stresses", series, name=name,
                         metadata=metadata, overwrite=overwrite)

    def add_model(self, ml: Union[ps.Model, dict],
                  overwrite: bool = False) -> None:
        """Add model to the database.

        Parameters
        ----------
        ml : pastas.Model or dict
            pastas Model or dictionary to add to the database
        overwrite : bool, optional
            if True, overwrite existing model, by default False

        Raises
        ------
        Exception
            if overwrite is False and model is already in the database
            raises an Exception.
        """
        lib = self.get_library("models")
        if ml.name not in lib.list_symbols() or overwrite:
            if isinstance(ml, ps.Model):
                mldict = ml.to_dict(series=False)
                name = ml.name
                metadata = ml.oseries.metadata
            elif isinstance(ml, dict):
                mldict = ml
                name = ml["name"]
                metadata = None
            else:
                raise TypeError("Expected pastas.Model or dict!")
            # check if oseries and stresses exist in store, if not add them
            self._check_model_series_for_store(ml)
            self._check_oseries_in_store(ml)
            self._check_stresses_in_store(ml)
            # write model to store
            lib.write(name, mldict, metadata=metadata)
        else:
            raise Exception("Model with name '{}' already in store!".format(
                ml.name))
        self._clear_cache("models")

    def _del_item(self, libname: str, name: str) -> None:
        """Internal method to delete items (series or models).

        Parameters
        ----------
        libname : str
            name of library to delete item from
        name : str
            name of item to delete
        """
        lib = self.get_library(libname)
        lib.delete(name)

    def del_models(self, names: Union[list, str]) -> None:
        """Delete model(s) from the database.

        Parameters
        ----------
        names : str or list of str
            name(s) of the model to delete
        """
        for n in self._parse_names(names, libname="models"):
            self._del_item("models", n)
        self._clear_cache("models")

    def del_oseries(self, names: Union[list, str]):
        """Delete oseries from the database.

        Parameters
        ----------
        names : str or list of str
            name(s) of the oseries to delete
        """
        for n in self._parse_names(names, libname="oseries"):
            self._del_item("oseries", n)
        self._clear_cache("oseries")

    def del_stress(self, names: Union[list, str]):
        """Delete stress from the database.

        Parameters
        ----------
        names : str or list of str
            name(s) of the stress to delete
        """
        for n in self._parse_names(names, libname="stresses"):
            self._del_item("stresses", n)
        self._clear_cache("stresses")

    def _get_series(self, libname: str, names: Union[list, str],
                    progressbar: bool = True) -> FrameorSeriesUnion:
        """Internal method to get timeseries.

        Parameters
        ----------
        libname : str
            name of the library
        names : str or list of str
            names of the timeseries to load
        progressbar : bool, optional
            show progressbar, by default True

        Returns
        -------
        pandas.DataFrame or dict of pandas.DataFrames
            either returns timeseries as pandas.DataFrame or
            dictionary containing the timeseries.
        """
        lib = self.get_library(libname)

        ts = {}
        names = self._parse_names(names, libname=libname)
        desc = f"Get {libname}"
        for n in (tqdm(names, desc=desc) if progressbar else names):
            ts[n] = lib.read(n).data
        # return frame if len == 1
        if len(ts) == 1:
            return ts[n]
        else:
            return ts

    def get_metadata(self, libname: str, names: Union[list, str],
                     progressbar: bool = False, as_frame: bool = True) -> \
            Union[dict, pd.DataFrame]:
        """Read metadata from database.

        Parameters
        ----------
        libname : str
            name of the library containing the dataset
        names : str or list of str
            names of the datasets for which to read the metadata

        Returns
        -------
        dict or pandas.DataFrame
            returns metadata dictionary (for one item) or DataFrame
            of metadata (for several datasets)
        """
        lib = self.get_library(libname)

        metalist = []
        names = self._parse_names(names, libname=libname)
        desc = f"Get metadata {libname}"
        for n in (tqdm(names, desc=desc) if progressbar else names):
            imeta = lib.read_metadata(n).metadata
            if imeta is None:
                imeta = {}
            if "name" not in imeta.keys():
                imeta["name"] = n
            metalist.append(imeta)
        if as_frame:
            meta = self._meta_list_to_frame(metalist, names=names)
            return meta
        else:
            if len(metalist) == 1:
                return metalist[0]
            else:
                return metalist

    def get_oseries(self, names: Union[list, str],
                    progressbar: bool = False) -> FrameorSeriesUnion:
        """Get oseries from database.

        Parameters
        ----------
        names : str or list of str
            names of the oseries to load
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        pandas.DataFrame or dict of DataFrames
            returns timeseries as DataFrame or dictionary of DataFrames if
            multiple names were passed
        """
        return self._get_series("oseries", names, progressbar=progressbar)

    def get_stresses(self, names: Union[list, str],
                     progressbar: bool = False) -> FrameorSeriesUnion:
        """Get stresses from database.

        Parameters
        ----------
        names : str or list of str
            names of the stresses to load
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        pandas.DataFrame or dict of DataFrames
            returns timeseries as DataFrame or dictionary of DataFrames if
            multiple names were passed
        """
        return self._get_series("stresses", names, progressbar=progressbar)

    def get_models(self, names: Union[list, str], return_dict: bool = False,
                   progressbar: bool = False) -> Union[ps.Model, dict]:
        """Load models from database.

        Parameters
        ----------
        names : str or list of str
            names of the models to load
        return_dict : bool, optional
            return model dictionary instead of pastas.Model (much
            faster for obtaining parameters, for example)
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        pastas.Model or list of pastas.Model
            return pastas model, or list of models if multiple names were
            passed
        """
        lib = self.get_library("models")
        models = []
        names = self._parse_names(names, libname="models")

        desc = "Get models"
        for n in (tqdm(names, desc=desc) if progressbar else names):
            item = lib.read(n)
            data = item.data
            if return_dict:
                ml = item.data
            else:
                ml = self._parse_model_dict(data)
            models.append(ml)
        if len(models) == 1:
            return models[0]
        else:
            return models

    @staticmethod
    def _clear_cache(libname: str) -> None:
        """Clear cached property."""
        getattr(ArcticConnector, libname).fget.cache_clear()

    @property  # type: ignore
    @functools.lru_cache()
    def oseries(self):
        """Dataframe with overview of oseries."""
        lib = self.get_library("oseries")
        df = self.get_metadata("oseries", lib.list_symbols())
        return df

    @property  # type: ignore
    @functools.lru_cache()
    def stresses(self):
        """Dataframe with overview of stresses."""
        lib = self.get_library("stresses")
        return self.get_metadata("stresses",
                                 lib.list_symbols())

    @property  # type: ignore
    @functools.lru_cache()
    def models(self):
        """List of model names."""
        lib = self.get_library("models")
        return lib.list_symbols()


class PystoreConnector(BaseConnector, ConnectorUtil):
    """Object to serve as the interface between storage and Python using the
    Pystore module. Provides all the methods to read, write, or delete data
    from the pystore.

    Create as PystoreConnector object that connects to a folder on disk
    containing a Pystore.

    Parameters
    ----------
    name : str
        name of the store
    path : str
        path to the pystore directory
    library_map: dict, optional
        dictionary containing the default library names as
        keys ('oseries', 'stresses', 'models') and the user
        specified library names as corresponding values.
        Allows user defined library names.
    """
    conn_type = "pystore"

    def __init__(self, name: str, path: str,
                 library_map: Optional[dict] = None):
        """Create a PystoreConnector object that points to a Pystore.

        Parameters
        ----------
        name : str
            name of the store
        path : str
            path to the pystore directory
        library_map : dict, optional
            dictionary containing the default library names as
            keys ('oseries', 'stresses', 'models') and the user
            specified library names as corresponding values.
            Allows user defined library names.
        """
        try:
            import pystore
        except ModuleNotFoundError as e:
            print("Install pystore, follow instructions at "
                  "https://github.com/ranaroussi/pystore#dependencies")
            raise e
        self.name = name
        self.path = path
        pystore.set_path(self.path)
        self.store = pystore.store(self.name)
        self.libs: dict = {}
        self._initialize(library_map)

    def __repr__(self):
        """Representation string of the object."""
        storename = self.name
        noseries = len(self.get_library("oseries").list_items())
        nstresses = len(self.get_library("stresses").list_items())
        nmodels = len(self.get_library("models").list_items())
        return (f"<PystoreConnector object> '{storename}': {noseries} oseries,"
                f" {nstresses} stresses, {nmodels} models")

    def _initialize(self, library_map: Optional[dict]):
        """Internal method to initalize the libraries (stores)."""
        if library_map is None:
            self.library_map = {i: i for i in self._default_library_names}
        else:
            self.library_map = library_map

        for libname in self.library_map.values():
            lib = self.store.collection(libname)
            self.libs[libname] = lib

    def get_library(self, libname: str):
        """Get Pystore library handle.

        Parameters
        ----------
        libname : str
            name of the library

        Returns
        -------
        Pystore.Collection handle
            handle to the library
        """
        # get custom library name if necessary
        real_libname = self.library_map[libname]

        # get library handle
        lib = self.store.collection(real_libname)
        return lib

    def _add_series(self, libname: str, series: FrameorSeriesUnion, name: str,
                    metadata: Optional[dict] = None,
                    overwrite: bool = False):
        """Internal method to add series to a library/store.

        Parameters
        ----------
        libname : str
            name of the library
        series : pandas.DataFrame or pandas.Series
            data to write to the pystore
        name : str
            name of the series
        metadata : dict, optional
            dictionary containing metadata, by default None
        overwrite : bool, optional
            overwrite existing dataset with the same name,
            by default False
        """
        self._validate_input_series(series)
        series = self._set_series_name(series, name)
        # convert to DataFrame because pystore doesn't accept pandas.Series
        # (maybe has an easy fix, but converting w to_frame for now)
        if isinstance(series, pd.Series):
            s = series.to_frame(name=name)
            is_series = True
        else:
            s = series
            is_series = False
        # store info about input series to ensure same type is returned
        if metadata is None:
            metadata = {"_is_series": is_series}
        else:
            metadata["_is_series"] = is_series
        lib = self.get_library(libname)
        if name not in lib.items or overwrite:
            lib.write(name, s, metadata=metadata, overwrite=overwrite)
        else:
            Exception("Item with name '{0}' already"
                      " in '{1}' library!".format(name, libname))
        self._clear_cache(libname)

    def add_oseries(self, series: FrameorSeriesUnion, name: str,
                    metadata: Optional[dict] = None,
                    overwrite=True):
        """Add oseries to the pystore.

        Parameters
        ----------
        series : pandas.DataFrame of pandas.Series
            oseries data to write to the store
        name : str
            name of series
        metadata : dict, optional
            dictionary containing metadata, by default None
        overwrite : bool, optional
            overwrite existing dataset with the same name,
            by default True
        """
        self._add_series("oseries", series, name,
                         metadata=metadata, overwrite=overwrite)

    def add_stress(self, series: FrameorSeriesUnion, name: str, kind: str,
                   metadata: Optional[dict] = None, overwrite=True):
        """Add stresses to the pystore.

        Parameters
        ----------
        series : pandas.DataFrame of pandas.Series
            stress data to write to the store
        kind : str
            category to identify type of stress, this label is added to the
            metadata dictionary.
        metadata : dict, optional
            dictionary containing metadata, by default None
        overwrite : bool, optional
            overwrite existing dataset with the same name,
            by default True
        """
        if metadata is None:
            metadata = {}
        if kind not in metadata.keys():
            metadata["kind"] = kind
        self._add_series("stresses", series, name,
                         metadata=metadata, overwrite=overwrite)

    def add_model(self, ml: Union[ps.Model, dict], overwrite: bool = True):
        """Add model to the pystore.

        Parameters
        ----------
        ml : pastas.Model or dict
            model to write to the store
        overwrite : bool, optional
            overwrite existing store model if it already exists,
            by default True
        """
        if isinstance(ml, ps.Model):
            mldict = ml.to_dict(series=False)
            name = ml.name
        elif isinstance(ml, dict):
            mldict = ml
            name = ml["name"]
        else:
            raise TypeError("Expected ps.Model or dict!")
        jsondict = json.loads(json.dumps(mldict, cls=PastasEncoder, indent=4))
        lib = self.get_library("models")
        # check if oseries and stresses exist in store, if not add them
        if not name in lib.items or overwrite:
            self._check_model_series_for_store(ml)
            self._check_oseries_in_store(ml)
            self._check_stresses_in_store(ml)
            lib.write(name, pd.DataFrame(), metadata=jsondict,
                      overwrite=overwrite)
        else:
            raise Exception("Model with name '{}' already in store!".format(
                name))
        self._clear_cache("models")

    def _del_series(self, libname: str, name):
        """Internal method to delete data from the store.

        Parameters
        ----------
        libname : str
            name of the library
        name : str
            name of the series to delete
        """
        lib = self.get_library(libname)
        lib.delete_item(name)
        self._clear_cache(libname)

    def del_oseries(self, names: Union[list, str]):
        """Delete oseries from pystore.

        Parameters
        ----------
        name : str
            name of the collection containing the data
        names : str or list of str, optional
            name(s) of oseries to delete
        """
        for n in self._parse_names(names, libname="oseries"):
            self._del_series("oseries", n)

    def del_stress(self, names: Union[list, str]):
        """Delete stresses from pystore.

        Parameters
        ----------
        names : str or list of str
            name(s) of the series to delete
        """
        for n in self._parse_names(names, libname="stresses"):
            self._del_series("stresses", n)

    def del_models(self, names: Union[list, str]):
        """Delete model(s) from pystore.

        Parameters
        ----------
        names : str
            name(s) of the model(s) to delete
        """
        for n in self._parse_names(names, libname="models"):
            self._del_series("models", n)

    def _get_series(self, libname: str, names: Union[list, str],
                    progressbar: bool = True):
        """Internal method to load timeseries data.

        Parameters
        ----------
        libname : str
            name of the store to load data from
        names : str or list of str
            name(s) of the timeseries to load
        progressbar : bool, optional
            show progressbar, by default True

        Returns
        -------
        pandas.DataFrame or dict of pandas.DataFrames
            returns data DataFrame or dictionary of DataFrames
            if multiple names are provided
        """
        lib = self.get_library(libname)

        ts = {}
        names = self._parse_names(names, libname=libname)
        desc = f"Get {libname}"
        for n in (tqdm(names, desc=desc) if progressbar else names):
            item = lib.item(n)
            s = item.to_pandas()
            # return pd.Series if user passed in Series
            is_series = item.metadata.pop("_is_series")
            if is_series:
                s = s.squeeze()
            ts[n] = s
        # return frame if only 1 name
        if len(ts) == 1:
            return ts[n]
        else:
            return ts

    def get_metadata(self, libname: str, names: Union[list, str],
                     progressbar: bool = False, as_frame=True) \
            -> Union[dict, pd.DataFrame]:
        """Read metadata from pystore.

        Parameters
        ----------
        libname : str
            name of the library the series are in,
            usually ("oseries" or "stresses")
        names : str or list of str
            name(s) of series to load metadata for
        progressbar : bool, optional
            show progressbar, by default True
        as_frame : bool, optional
            return metadata as dataframe, default is
            True, otherwise return as dict or list of
            dict

        Returns
        -------
        list or pandas.DataFrame
            list or pandas.DataFrame containing metadata
        """
        import pystore
        lib = self.get_library(libname)

        metalist = []
        names = self._parse_names(names, libname=libname)
        desc = f"Get metadata {libname}"
        for n in (tqdm(names, desc=desc) if progressbar else names):
            imeta = pystore.utils.read_metadata(lib._item_path(n))
            if "name" not in imeta.keys():
                imeta["name"] = n
            if "_is_series" in imeta.keys():
                imeta.pop("_is_series")
            metalist.append(imeta)
        if as_frame:
            meta = self._meta_list_to_frame(metalist, names=names)
            return meta
        else:
            if len(metalist) == 1:
                return metalist[0]
            else:
                return metalist

    def get_oseries(self, names: Union[list, str],
                    progressbar: bool = False) -> FrameorSeriesUnion:
        """Retrieve oseries from pystore.

        Parameters
        ----------
        names : str or list of str
            name(s) of collections to load oseries data from
        progressbar : bool, optional
            show progressbar, by default True

        Returns
        -------
        pandas.DataFrame or dict of pandas.DataFrames
            returns data as a DataFrame or a dictionary of DataFrames
            if multiple names are passed
        """
        return self._get_series("oseries", names, progressbar=progressbar)

    def get_stresses(self, names: Union[list, str],
                     progressbar: bool = False) -> FrameorSeriesUnion:
        """Retrieve stresses from pystore.

        Parameters
        ----------
        names : str or list of str
            name(s) of collections to load stresses data from
        progressbar : bool, optional
            show progressbar, by default True

        Returns
        -------
        pandas.DataFrame or dict of pandas.DataFrames
            returns data as a DataFrame or a dictionary of DataFrames
            if multiple names are passed
        """
        return self._get_series("stresses", names, progressbar=progressbar)

    def get_models(self, names: Union[list, str], return_dict: bool = False,
                   progressbar: bool = False) -> Union[ps.Model, dict]:
        """Load models from pystore.

        Parameters
        ----------
        names : str or list of str
            name(s) of the models to load
        return_dict : bool, optional
            return model dictionary instead of pastas.Model object
            (much faster for obtaining parameters, for example)
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        list or pastas.Model
            model or list of models
        """
        lib = self.get_library("models")

        models = []
        load_mod = import_module("pastas.io.pas")  # type: ignore
        names = self._parse_names(names, libname="models")

        desc = "Get models"
        for n in (tqdm(names, desc=desc) if progressbar else names):
            jsonpath = lib._item_path(n).joinpath("metadata.json")
            data = load_mod.load(jsonpath)  # type: ignore
            if return_dict:
                ml = data
            else:
                ml = self._parse_model_dict(data)
            models.append(ml)
        if len(models) == 1:
            return models[0]
        else:
            return models

    @staticmethod
    def _clear_cache(libname: str) -> None:
        """Clear cached property."""
        getattr(PystoreConnector, libname).fget.cache_clear()

    @property  # type: ignore
    @functools.lru_cache()
    def oseries(self):
        """Dataframe with overview of oseries."""
        lib = self.get_library("oseries")
        df = self.get_metadata("oseries", lib.list_items())
        return df

    @property  # type: ignore
    @functools.lru_cache()
    def stresses(self):
        """Dataframe with overview of stresses."""
        lib = self.get_library("stresses")
        df = self.get_metadata("stresses", lib.list_items())
        return df

    @property  # type: ignore
    @functools.lru_cache()
    def models(self):
        """List of model names."""
        lib = self.get_library("models")
        if lib is not None:
            mls = lib.list_items()
        else:
            mls = []
        return mls


class DictConnector(BaseConnector, ConnectorUtil):
    """Object to store timeseries and pastas models in-memory. Provides methods
    to read, write, or delete data from the object. Data is stored in
    dictionaries.

    Parameters
    ----------
    name : str
        user-specified name of the connector
    library_map : dict, optional
        dictionary containing the default library names as
        keys ('oseries', 'stresses', 'models') and the user
        specified library names as corresponding values.
        Allows user defined library names.
    """
    conn_type = "dict"

    def __init__(self, name: str, library_map: Optional[dict] = None):
        """Create DictConnector object that stores data in dictionaries.

        Parameters
        ----------
        name : str
            user-specified name of the connector
        library_map : dict, optional
            dictionary containing the default library names as
            keys ('oseries', 'stresses', 'models') and the user
            specified library names as corresponding values.
            Allows user defined library names.
        """
        self.name = name

        # allow custom library names
        if library_map is None:
            libmap = {i: i for i in self._default_library_names}
        else:
            libmap = library_map

        self.library_map = libmap

        # set empty dictionaries for series
        for val in self.library_map.values():
            setattr(self, "lib_" + val, {})

    def __repr__(self):
        """Representation string of the object."""
        noseries = len(self.get_library("oseries").keys())
        nstresses = len(self.get_library("stresses").keys())
        nmodels = len(self.get_library("models").keys())
        return ("<DictConnector object> '{0}': {1} oseries, {2} stresses, "
                "{3} models".format(self.name, noseries, nstresses, nmodels))

    def get_library(self, libname: str):
        """Get reference to dictionary holding data.

        Parameters
        ----------
        libname : str
            name of the library
        """
        # get custom library name
        real_libname = "lib_" + self.library_map[libname]
        return getattr(self, real_libname)

    def _add_series(self, libname: str, series: FrameorSeriesUnion,
                    name: str, metadata: Union[dict, None] = None,
                    overwrite: bool = False) -> None:
        """Internal method to obtain series.

        Parameters
        ----------
        libname : str
            name of library
        series : FrameorSeriesUnion
            pandas.Series or pandas.DataFrame containing data
        name : str
            name of the series
        metadata : dict, optional
            dictionary containing metadata, by default None
        overwrite : bool, optional
            overwrite existing dataset with the same name,
            by default False
        """
        self._validate_input_series(series)
        series = self._set_series_name(series, name)
        lib = self.get_library(libname)
        if name not in lib or overwrite:
            lib[name] = (metadata, series)
        else:
            Exception("Item with name '{0}' already"
                      " in '{1}' library!".format(name, libname))
        self._clear_cache(libname)

    def add_oseries(self, series: FrameorSeriesUnion, name: str,
                    metadata: Union[dict, None] = None,
                    overwrite: bool = False) -> None:
        """Add oseries to object.

        Parameters
        ----------
        series : FrameorSeriesUnion
            pandas.Series or pandas.DataFrame containing data
        name : str
            name of the oseries
        metadata : dict, optional
            dictionary with metadata, by default None
        overwrite : bool, optional
            overwrite existing timeseries, default is False
        """
        self._add_series("oseries", series, name, metadata=metadata,
                         overwrite=overwrite)

    def add_stress(self, series: FrameorSeriesUnion, name: str, kind: str,
                   metadata: Union[dict, None] = None,
                   overwrite: bool = False) -> None:
        """Add stress to object.

        Parameters
        ----------
        series : FrameorSeriesUnion
            pandas.Series or pandas.DataFrame containing data
        name : str
            name of the stress
        kind : str
            type of stress (i.e. 'prec', 'evap', 'well', etc.)
        metadata : dict, optional
            dictionary containing metadata, by default None
        overwrite : bool, optional
            overwrite existing timeseries, default is False
        """
        if metadata is None:
            metadata = {}
        if kind not in metadata.keys():
            metadata["kind"] = kind
        self._add_series("stresses", series, name, metadata=metadata,
                         overwrite=overwrite)

    def add_model(self, ml: Union[ps.Model, dict],
                  overwrite: bool = False) -> None:
        """Add model to object.

        Parameters
        ----------
        ml : pastas.Model or dict
            pastas.Model or dictionary to add
        overwrite : bool, optional
            overwrite model in store, default is False
        """
        lib = self.get_library("models")
        if isinstance(ml, ps.Model):
            mldict = ml.to_dict(series=False)
            name = ml.name
        elif isinstance(ml, dict):
            mldict = ml
            name = ml["name"]
        else:
            raise TypeError("Expected pastas.Model or dict!")
        # check if oseries and stresses exist in store, if not add them
        if name not in lib or overwrite:
            self._check_model_series_for_store(ml)
            self._check_oseries_in_store(ml)
            self._check_stresses_in_store(ml)
            lib[name] = mldict
        else:
            raise Exception("Model with name '{}' already in store!".format(
                name))
        self._clear_cache("models")

    def del_models(self, names: Union[list, str]) -> None:
        """Delete models from object.

        Parameters
        ----------
        names : Union[list, str]
            str or list of str of model names to remove
        """
        lib = self.get_library("models")
        for n in self._parse_names(names, libname="models"):
            _ = lib.pop(n)
        self._clear_cache("models")

    def del_oseries(self, names: Union[list, str]) -> None:
        """Delete oseries from object.

        Parameters
        ----------
        names : Union[list, str]
            str or list of str of oseries to remove
        """
        lib = self.get_library("oseries")
        for n in self._parse_names(names, libname="oseries"):
            _ = lib.pop(n)
        self._clear_cache("oseries")

    def del_stress(self, names: Union[list, str]) -> None:
        """Delete stresses from object.

        Parameters
        ----------
        names : Union[list, str]
            str or list of str of stresses to remove
        """
        lib = self.get_library("stresses")
        for n in self._parse_names(names, libname="stresses"):
            _ = lib.pop(n)
        self._clear_cache("stresses")

    def _get_series(self, libname: str, names: Union[list, str],
                    progressbar: bool = True) -> FrameorSeriesUnion:
        """Internal method to get oseries or stresses.

        Parameters
        ----------
        libname : str
            name of library
        names : Union[list, str]
            str or list of string
        progressbar : bool, optional
            show progressbar, by default True

        Returns
        -------
        dict, FrameorSeriesUnion
            returns DataFrame or Series if only one name is passed, else
            returns dict with all the data
        """
        lib = self.get_library(libname)
        ts = {}
        names = self._parse_names(names, libname=libname)
        desc = f"Get {libname}"
        for n in (tqdm(names, desc=desc) if progressbar else names):
            ts[n] = deepcopy(lib[n][1])
        # return frame if len == 1
        if len(ts) == 1:
            return ts[n]
        else:
            return ts

    def get_metadata(self, libname: str, names: Union[list, str],
                     progressbar: bool = False, as_frame: bool = True) \
            -> Union[pd.DataFrame, list]:
        """Get metadata from object.

        Parameters
        ----------
        libname : str
            name of library
        names : Union[list, str]
            str or list of str of names to get metadata for
        progressbar : bool, optional
            show progressbar, by default False
        as_frame : bool, optional
            return as DataFrame, by default True

        Returns
        -------
        Union[pd.DataFrame, list]
            returns list of metadata or pandas.DataFrame depending on value
            of `as_frame`
        """
        lib = self.get_library(libname)
        metalist = []
        names = self._parse_names(names, libname=libname)
        desc = f"Get metadata {libname}"
        for n in (tqdm(names, desc=desc) if progressbar else names):
            imeta = deepcopy(lib[n][0])
            if imeta is None:
                imeta = {}
            if "name" not in imeta.keys():
                imeta["name"] = n
            metalist.append(imeta)

        if as_frame:
            meta = self._meta_list_to_frame(metalist, names=names)
            return meta
        else:
            if len(metalist) == 1:
                return metalist[0]
            else:
                return metalist

    def get_oseries(self, names: Union[list, str],
                    progressbar: bool = False) -> FrameorSeriesUnion:
        """Retrieve oseries from object.

        Parameters
        ----------
        names : Union[list, str]
            name or list of names to retrieve
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        dict, FrameorSeriesUnion
            returns dictionary or DataFrame/Series depending on number of
            names passed
        """
        return self._get_series("oseries", names, progressbar=progressbar)

    def get_stresses(self, names: Union[list, str],
                     progressbar: bool = False) -> FrameorSeriesUnion:
        """Retrieve stresses from object.

        Parameters
        ----------
        names : Union[list, str]
            name or list of names of stresses to retrieve
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        dict, FrameorSeriesUnion
            returns dictionary or DataFrame/Series depending on number of
            names passed
        """
        return self._get_series("stresses", names, progressbar=progressbar)

    def get_models(self, names: Union[list, str], return_dict: bool = False,
                   progressbar: bool = False) -> Union[Model, dict]:
        """Load models from object.

        Parameters
        ----------
        names : str or list of str
            names of the models to load
        return_dict : bool, optional
            return model dictionary instead of pastas.Model object
            (much faster for obtaining parameters, for example)
        progressbar : bool, optional
            show progressbar, by default False

        Returns
        -------
        pastas.Model or list of pastas.Model
            return pastas model, or list of models if multiple names were
            passed
        """
        lib = self.get_library("models")
        models = []
        names = self._parse_names(names, libname="models")

        desc = "Get models"
        for n in (tqdm(names, desc=desc) if progressbar else names):
            data = deepcopy(lib[n])
            if return_dict:
                ml = data
            else:
                ml = self._parse_model_dict(data)
            models.append(ml)
        if len(models) == 1:
            return models[0]
        else:
            return models

    @staticmethod
    def _clear_cache(libname: str) -> None:
        """Clear cached property."""
        getattr(DictConnector, libname).fget.cache_clear()

    @property  # type: ignore
    @functools.lru_cache()
    def oseries(self):
        """Dataframe showing overview of oseries."""
        lib = self.get_library("oseries")
        return self.get_metadata("oseries", names=list(lib.keys()))

    @property  # type: ignore
    @functools.lru_cache()
    def stresses(self):
        """Dataframe showing overview of stresses."""
        lib = self.get_library("stresses")
        return self.get_metadata("stresses", names=list(lib.keys()))

    @property  # type: ignore
    @functools.lru_cache()
    def models(self):
        """List of model names."""
        lib = self.get_library("models")
        return list(lib.keys())



from setuptools import setup, find_packages

# try:
#     import pypandoc
#     l_d = pypandoc.convert('README.md', )
# except ModuleNotFoundError:
l_d = ''

# Get the version.
version: dict = {}
with open("version.py") as fp:
    exec(fp.read(), version)
setup(
    name='pastastore',
    version=version['__version__'],
    description='pastastore module by Artesia',
    long_description=l_d,
    url='https://artesia-water.nl',
    author='Artesia',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    platforms='Windows, Mac OS-X',
    install_requires=['numpy>=1.15', 'pandas>=0.24', "tqdm",
                      "pastas>=0.13", "arctic"],
    packages=find_packages(exclude=[]),
    include_package_data=True,
)
